;;; -*- Mode: Emacs-Lisp; outline-regexp: "\n;;;;+" -*-

;;;;;; Lisp-Utils: Emacs extensions for working with Common Lisp.
;;;;;; Version 1
;;;;;; http://asymmetrical-view.com/personal/code/emacs/kburton/lisp-utils.el

;;; This module was written by Kyle R. Burton <kyle.burton@gmail.com>
;;; (exept where explicitly noted) and placed in the Public Domain.
;;; All warranties are disclaimed.

;;; Special thanks and mention goes to Taylor R. Campbell and his
;;; paredit.el[2] module, upon which this module is both inspired by as
;;; well as depends on.

;;; The functions and keybindings in this module provide a layer above
;;; and beyond both Emac's lisp-mode and paredit-mode.  They create a
;;; more IDE-like experience when working with Common Lisp code.
;;; Where paredit elimiates most possible mistakes with strings and
;;; sexprs (it in fact eliminates the posibility of unbalanced
;;; parentheses), this module assists with the authoring of many of
;;; the common forms themselves (package declarations, exports, defuns
;;; and so on).

;;; Many of my personal conventions are in fact embodied by the
;;; functions in this module, my personal preference for the
;;; lisp-unit[1] unit testing framework, my preference for creating
;;; asf packages and other naming and code organization preferences.
;;; Attempts have been made to factor out some of these
;;; personalizations that others may wish to override with their own
;;; behavior and preferences, though there are doubtless areas you may
;;; find not to your liking - I'm very open to submissions and
;;; patches.

;;; [1] lisp-unit: http://www.cs.northwestern.edu/academics/courses/325/readings/lisp-unit.html
;;; [2] paredit: http://mumble.net/~campbell/emacs/paredit.el

;;; TODO
;;;  * Make krb-lisp-export-symbol idempotent (it should keep the
;;;    exports list organized - uniquely sorted?
;;;  * refactor-selection-as-defun
;;;  * copy-defun-body-to-point
;;;  * make insert-defun realize if a function already exists, and just
;;;    jump to it's definition
;;;  * this leads to having a \C-xp. and \C-xp* equivalent to a
;;;    tag-stack, see perl-utils.el for the example
;;;  * implement add-use-library (adds to the defpackage's (:use...)
;;;    list, should also be idempotent and organize the use list

;; This module absolutely requires paredit
(require 'paredit)
(require 'cl)

(add-hook
 'lisp-mode-hook
 (lambda ()
   (paredit-mode +1)
   (krb-lisp-apply-keybindings)))

(add-hook
 'emacs-lisp-mode-hook
 (lambda ()
   (paredit-mode +1)
   (krb-lisp-apply-keybindings)))

;; this is here so these macros can maintain the defvars...
;; (cl:defpackage ...)

(defvar *krb-lisp-unit-test-cmdline-func* 'krb-lisp-make-test-runner-cmdline
  "Generates the unit test runner command line.")

(defvar *krb-lisp-project-directory* (concat (getenv "HOME") "/personal/projects")
  "The base location of your projects.  You will probably want to
override this in your .emacs after you've loaded lisp-utils.el,
see also `*krb-lisp-lisp-project-directory*'")

(defvar *krb-lisp-lisp-project-directory* (concat *krb-lisp-project-directory* "/lisp")
  "The location (relative to `*krb-lisp-project-directory*') of
your lisp projects.  You will probably want to override this in
your .emacs after you've loaded lisp-utils.el.")

(defvar *krb-lisp-keybinding-prefix* "\C-xp"
  "This is a prefix for all of the keybindings used by this
  module.  See also `*krb-lisp-keybinding-table*', which you may
  want to customize to suit your preferences.  There is no
  special mnemonic for this, it was just a key prefix that wasn't
  commonly used by other functions.")

(defvar *krb-lisp-keybinding-table* (list))

(defvar *krb-lisp-find-defvar-insertion-point-function*
  'krb-lisp-move-to-defvar-insertion-point
  "This is used for determining where you want your defvar's
  maintained.  The default is
  `krb-lisp-move-to-defvar-insertion-point', which will place
  defvar forms just after the defpackage form")

(defvar *krb-lisp-find-defun-insertion-point-function*
  'end-of-buffer
  "This is used for determining where you want your defun's and
  defmacro's maintained.  The default is `end-of-buffer'.")

(defvar *krb-lisp-jump-stack* (list)
  "Used by the '.' and '*' (pop) keybindings as a tag-like stack.")

(setf *krb-lisp-keybinding-table*
  '(("a"     krb-lisp-new-asdf-system)
    ("e"     krb-lisp-export-symbol) 
    ("f"     krb-lisp-insert-function)
    ("p"     krb-lisp-insert-defpackage)
    ("t"     krb-lisp-switch-to-test-for-defun)
    ("T"     krb-lisp-switch-buffers-with-test)
    ("\C-t"  krb-lisp-execute-unit-tests-for-buffer)
    ("v"     krb-lisp-insert-defvar)
    (")"     paredit-forward-slurp-sexp)
    ("("     paredit-backward-slurp-sexp)
    ("}"     paredit-forward-barf-sexp)
    ("{"     paredit-backward-barf-sexp)
    ("."     krb-lisp-jump-to-def-form-push)
    ("*"     krb-lisp-pop-jump-stack)
    (">"     krb-lisp-find-next-symbol-usage)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; functions

(defun krb-lisp-apply-keybindings ()
  "Uses local-set-key to apply keybindings"
  (interactive)
  (loop for (binding fname) in *krb-lisp-keybinding-table*
	do
	(local-set-key (concat *krb-lisp-keybinding-prefix* binding)
		       fname)))

(defun krb-lisp-insert-indented (&rest strings)
  (loop for string in strings
	do
	(insert string)
	(indent-for-tab-command)))

(defun krb-lisp-to-string (thing)
  "converts the thing to a strin"
  (format "%s" thing))

(defun krb-lisp-insert-indented-lines (&rest lines)
  "Loops over the given list of items, converting them to
strings, inserts the item with a trailing newline and then
invokes `indent-for-tab-command'."
  (loop for item in lines
	do
	(cond ((listp item)
	       (krb-lisp-insert-indented (apply #'concat item)))
	      ((stringp item)
	       (krb-lisp-insert-indented (concat item "\n")))
	      (t
	       (error "Don't know how to stringify: %s" item)))))

(defun krb-lisp-intersperse (delim things)
  (labels ((isp (things acc)
	     (cond ((null things)
		  acc)
		 (t
		  (isp (cdr things)
		       (cons (car things) (cons delim
					  acc)))))))
    (cond ((null things)
	   nil)
	  (t
	   (reverse (isp (cdr things)
		  (list (car things))))))))

(defun krb-lisp-flatten (&rest trees)
  (cond ((null trees)
	 (list))
	((listp (car trees))
	 (append
	  (apply #'krb-lisp-flatten (car trees))
	  (apply #'krb-lisp-flatten (cdr trees))))
	(t
	 (cons (car trees)
	       (apply #'krb-lisp-flatten (cdr trees))))))

(defun krb-lisp-join->string (delim &rest strings)
  (apply #'concat
	 (mapcar #'(lambda (elt)
		      (format "%s" elt))
		 (krb-lisp-intersperse
		  delim
		  (krb-lisp-flatten strings)))))

(defun krb-lisp-insert-defpackage (package-name use-list exports)
  (interactive
   (list
    (read-string "Package Name: "
		 (concat "#:"   
			 (file-name-sans-extension
			  (file-name-nondirectory
			   (buffer-file-name)))))
    (read-string "Use Packages: " '("#:common-lisp"))
    (read-string "Exports: " "")))
  (krb-lisp-insert-indented-lines
   `("(cl:defpackage " ,package-name "\n")
   `("(:use " ,use-list ")\n")
   `("(:export " ,exports "))\n")
   `("\n")
   `("(cl:in-package " ,package-name ")\n")))

(defun krb-lisp-move-to-defvar-insertion-point ()
  (krb-lisp-find-package-declaration)
  (search-forward-regexp "(\\(cl:\\)?in-package")
  (next-line 1))

(defun krb-lisp-insert-defvar (var-name value docstring)
  (interactive
   (list
    (read-string (format "Name [%s]: " (krb-lisp-symbol-at-cursor))  nil nil (krb-lisp-symbol-at-cursor))
    (read-string "Value [nil]: " nil nil "nil")
    (read-string "Doc: ")))
  (save-excursion
    (funcall *krb-lisp-find-defvar-insertion-point-function*)
    (insert "\n")
    (krb-lisp-insert-indented "(defvar " var-name " " value "\n")
    (krb-lisp-insert-indented "\"" docstring "\")\n")))


;; TODO: this doesn't work in emacs-lisp-mode buffers, can that be
;; accomodated as well as common-lisp?

(defun krb-lisp-export-symbol (symbol-name)
  "Finds the package declaration relative to the current
file (see krb-lisp-find-package-declaration), and adds
symbol-name to the export list."
  (interactive
   (list (read-string "Symbol: " (krb-lisp-find-likely-symbol))))
  (let ((orig-buffer (current-buffer)))
    (save-excursion
      (krb-lisp-find-package-declaration)
      (search-forward-regexp "(:?export")
      (search-backward "(")
      (forward-sexp 1)
      (backward-char 1)
      (insert "\n#:" symbol-name)
      (indent-for-tab-command)
      (save-buffer)
      (switch-to-buffer orig-buffer))))

(defun krb-lisp-find-likely-symbol ()
  (interactive)
  (condition-case err
      (save-excursion
	(cond
	 ((paredit-in-string-p)
	  (goto-char (car (paredit-string-start+end-points)))
	  (backward-char 2)
	  (krb-lisp-get-prev-defun-name))
	 (t
	  (let ((sym (krb-lisp-symbol-at-cursor)))
	    (if (string-match "[.)(]" sym)
		(krb-lisp-get-prev-defun-name)
	      sym)))))
    (error 
     (message "krb-lisp-find-likely-symbol: err:%s" err))))



(defun krb-lisp-get-prev-defun-name ()
  "Looks backwards for the enclosing defun, currently this
function assumes that you're inside a defun, if you're not, it'll
just find the previous defun's name."
  (interactive)
  (save-excursion
    (search-backward-regexp "(def\\(un\\|macro\\|method\\|var\\|parameter\\|class\\|struct\\) ")
    (forward-word 1)
    (forward-char 1)
    (let ((start (point)))
      (search-forward-regexp " ")
      (buffer-substring start (- (point) 1)))))

(defun krb-lisp-symbol-at-cursor ()
  "Looks for the most likely symbol name closest to where the
cursor is, with a right (forward) bias based on distance if the
word is not on a symbol character."
  (interactive)
  ;; pulled from etags.el in Emacs 21.1, noversion was on
  ;; etags.el, latest copyright was 2001
  (funcall (or find-tag-default-function
               (get major-mode 'find-tag-default-function)
               'find-tag-default)))


(defun krb-lisp-find-package-declaration ()
  "Finds a common-lisp package declaration, starting with the
  current buffer, then looking for a file called package.lisp
  starting in the same directory as the current file, moving
  backwards one directory at a time until such a file is located
  or the root of the file system is reached, upon which this
  function errors.

  The point will be placed at the beginning of the package
  declaration, if located."
  (interactive)
  (condition-case
      ()
      (krb-lisp-find-package-declaration-in-file)
    (error
     (progn
       (krb-lisp-find-file-by-name-traverse-up "package.lisp")
       (krb-lisp-find-package-declaration-in-file)))))

(defun krb-lisp-find-file-by-name-traverse-up (file-name &optional path)
  "Using the location of the current buffer as a default (unless
path, colon-seperated string of paths to search, is supplied),
looks for and opens a buffer for the given file name, which
should not have a path."
  (interactive "sFile Name: ")
  (unless path
    (setf path (replace-regexp-in-string "/$" "" (file-name-directory (buffer-file-name)))))
  (cond ((file-exists-p (concat path "/" file-name))
	 (find-file (concat path "/" file-name)))
	(t
	 (krb-lisp-find-file-by-name-traverse-up
	  file-name
	  (krb-lisp-strip-path-suffix path 1)))))

(defun krb-lisp-find-file-by-wildcard-traverse-up (file-pattern &optional path)
  "Using the location of the current buffer as a default (unless
path is given), looks for and returns the full file name for the first file matching the given file
pattern.  This function errors if more than 1 file matches the pattern."
  (interactive "sFile Name: ")
  (unless path
    (setf path (replace-regexp-in-string "/$" "" (file-name-directory (buffer-file-name)))))
  (let ((files (file-expand-wildcards (expand-file-name (concat path "/" file-pattern)))))
    (cond
     ((= 1 (length files))
      (car files))
     (t
      (krb-lisp-find-file-by-wildcard-traverse-up
       file-pattern
       (krb-lisp-strip-path-suffix path 1))))))

(defun krb-lisp-asdf-find-defsystem-file ()
  (interactive)
  (let ((file (find-file (krb-lisp-find-file-by-wildcard-traverse-up "*.asd"))))
    (message "asdf defsystem file: %s" file)
    file))

;; TODO: this is broken...
(defun krb-lisp-asdf-add-component (component-type component-name depenent-components)
  (interactive
   (list
    (read-string "Component Type [file]: " nil nil "file")
    (read-string (format "Component Name [%s]: "
			 (file-name-sans-extension (file-name-nondirectory (buffer-file-name))))
		 nil
		 nil
		 (file-name-sans-extension (file-name-nondirectory (buffer-file-name))))
    (read-string (format "Dependencies [%s]: " ""))))
  (save-excursion
    (find-file (krb-lisp-find-file-by-wildcard-traverse-up "*.asd"))
    (beginning-of-buffer)
    (search-forward ":components ")
    (forward-sexp 1)
    (backward-char 1)
    (krb-lisp-insert-indented "\n")
    (krb-lisp-insert-indented-lines
     `("(:" ,component-type " \"" ,component-name "\"\n")
     `(":depends-on (" ,depenent-components "))"))
    (save-buffer)
    (switch-to-buffer nil)))

(defun krb-lisp-strip-path-suffix (path parts)
  "Strips N (parts) path parts off of the given path name.

  (krb-lisp-strip-path-suffix \"/foo/bar/qux\" 1)
     => \"/foo/bar\" "
  (interactive "sPath: \nnParts: ")
  (cond ((= 0 parts)
	 path)
	(t
	 (krb-lisp-strip-path-suffix
	  (replace-regexp-in-string "/[^/]*$" "" path)
	  (- parts 1)))))


(defun krb-lisp-find-package-declaration-in-file ()
  (interactive)
  (goto-char (point-min))
  (let ((case-fold-search t))
    (search-forward-regexp "(\\(cl\\|common-lisp\\)?:?defpackage"))
  (backward-sexp 1))


(defun krb-lisp-find-likely-arglist ()
  "Pulls the cdr of the current form as the default, if within a
string, exits the string (on the right) and returns the cdr of
the following sexpr."
  (interactive)
  (save-excursion
    (cond ((paredit-in-string-p)
	   (goto-char (cdr (paredit-string-start+end-points)))
	   (search-forward "(")
	   (forward-char 1)
	   (forward-sexp 1)
	   (paredit-skip-whitespace t)
	   (let ((start (point)))
	     (search-forward ")")
	     (forward-char 1)
	     (buffer-substring start (point))))
	  (t
	   (search-backward "(")
	   (forward-char 1)
	   (forward-sexp 1)
	   (paredit-skip-whitespace t)
	   (let ((start (point)))
	     (search-forward ")")
	     (backward-char 1)
	     (buffer-substring start (point)))))))

(defun krb-tmp ()
  (interactive)
  (message " => '%s'" (krb-lisp-find-likely-arglist)))


;; if the function already exists in the current file, we should jump to it
(defun krb-lisp-insert-function (fn-name args)
  (interactive
   (list
    (read-string (format "Name [%s]: " (krb-lisp-find-likely-symbol))
		 nil nil (krb-lisp-find-likely-symbol))
    (read-string (format "Arguments [%s] (* for none): " (krb-lisp-find-likely-arglist))
		 nil nil (krb-lisp-find-likely-arglist))))
  (funcall *krb-lisp-find-defun-insertion-point-function*)
  (if (string= "*" args)
      (setf args ""))
  (krb-lisp-insert-indented "(defun " fn-name " (" args ")\n")
  (krb-lisp-insert-indented "\"The author needs to document this function\"\n")
  (krb-lisp-insert-indented "(error \"implement this\"))")
  (krb-lisp-insert-indented "\n")
  (search-backward "(error"))


(defun krb-lisp-switch-buffers-with-test ()
  "If in a buffer, for example named: foo.lisp, this function
will swtich to the corresponding unit test buffer (lisp-unit)
named test-foo.lisp.

If you're in the buffer test-foo.lisp, this function will swtich
to the buffer foo.lisp, opening the file if necessary.

In the first case, if the file test-foo.lisp does not exist, it
will create it, generating a basic unit test template."
  (interactive)
  (let ((file-name (krb-lisp-get-test-file-name-for-current-buffer)))
    (cond ((string= (buffer-file-name)
		    file-name)
	   (find-file (krb-lisp-get-base-file-name-for-current-test-buffer file-name)))
	  (t
	   (let ((need-to-generate (not (file-exists-p file-name)))
		 (base-package (krb-lisp-get-package-name-for-current-buffer)))
	     (find-file file-name)
	     (when need-to-generate
	       (krb-lisp-generate-test-package
		(file-name-sans-extension (file-name-nondirectory file-name))
		(format ":common-lisp :lisp-unit %s" base-package)
		"")))))))

(defun krb-lisp-find-test-buffer (file-name)
  "If in a buffer, for example named: foo.lisp, this function
will swtich to the corresponding unit test buffer (lisp-unit)
named test-foo.lisp.

If you're in the buffer test-foo.lisp, this function will swtich
to the buffer foo.lisp, opening the file if necessary.

In the first case, if the file test-foo.lisp does not exist, it
will creat it, generating a basic unit test template."
  (interactive
   (list (read-string (format "Test File Name [%s]: "
			      (krb-lisp-get-test-file-name-for-current-buffer))
		      nil
		      nil
		      (krb-lisp-get-test-file-name-for-current-buffer))))
  (cond ((string= (buffer-file-name)
		  file-name)
	 (find-file (krb-lisp-get-base-file-name-for-current-test-buffer file-name)))
	(t
	 (let ((need-to-generate (not (file-exists-p file-name)))
	       (base-package (krb-lisp-get-package-name-for-current-buffer)))
	   (find-file file-name)
	   (when need-to-generate
	     (krb-lisp-generate-test-package
	      (file-name-sans-extension (file-name-nondirectory file-name))
	      (format ":common-lisp :lisp-unit %s" base-package)
	      ""))))))

(defun krb-lisp-get-test-file-name-for-current-buffer ()
 "Computes an appropriate name for the corresponding unit test
file for the current buffer's file name.  Returns the name of the
current buffer if already in a test buffer (name starts with
test)."
  (cond ((string-match "/test-" (buffer-file-name))
	 (buffer-file-name))
	(t
	 (replace-regexp-in-string "/\\([^/]+\\)$"
				   "/test-\\1"
				   (buffer-file-name)))))

(defun krb-lisp-generate-test-package (package-name use-packages exports)
  "Inserts the package declaration and imports for a lisp-unit unit test."
  (interactive
   (list
    (read-string "Package Name: "
		 (concat "#:"
			 (file-name-sans-extension
			  (file-name-nondirectory
			   (buffer-file-name)))))
    (read-string "Use Packages: " '(":common-lisp :lisp-unit"))
    (read-string "Exports: " "")))
  (krb-lisp-insert-indented-lines
   `("(cl:defpackage " ,package-name "\n")
   `("(:use " ,use-packages ")\n")
   `("(:export " ,exports "))\n")
   `("\n")
   `("(cl:in-package " ,package-name ")\n")))

(defun krb-lisp-get-package-name-for-current-buffer ()
  "Finds or computes a package name for the current buffer."
  (file-name-sans-extension (file-name-nondirectory (buffer-file-name))))

(defun krb-lisp-get-test-package-name-for-current-buffer ()
  "Finds or computes a test package name for the current buffer."
  (let ((package-name (file-name-nondirectory (file-name-sans-extension (buffer-file-name)))))
    (cond ((string-match "^:test-" package-name)
	   (substring package-name 1))
	  ((string-match "^test-" package-name)
	   package-name)
	  (t
	   (concat "test-" package-name)))))

(defun krb-lisp-switch-to-test-for-defun (test-name)
  "Generates or jumps to the lisp-unit test definition for the
current defun, generating it (as well as the unit test) if
necessary."
  (interactive
   (list
    (read-string (format "Test Name [test-%s]: " (krb-lisp-get-prev-defun-name))
		 nil nil (krb-lisp-get-prev-defun-name))))
  (krb-lisp-switch-buffers-with-test)
  (end-of-buffer)
  (insert "\n")
  (krb-lisp-insert-indented (format "(define-test test-%s\n" test-name))
  (krb-lisp-insert-indented "(assert-true nil \"Please implement the test\"))\n")
  (search-backward "(assert-true"))


(defun krb-lisp-get-base-file-name-for-current-test-buffer (file-name)
  "Strips off the \"test-\" prefix on the file."
  (interactive
   (list (read-string (format "File Name [%s]: " (buffer-file-name)
			      nil nil (buffer-file-name)))))
  (replace-regexp-in-string "/test-" "/" file-name))


(defun krb-lisp-execute-unit-tests-for-buffer ()
  "Using lisp-unit, executes the unit test module for the current
buffer (computed via `krb-lisp-get-test-file-name-for-current-buffer')."
  (interactive)
  (message "running...")
  (save-excursion
    ;;    (switch-to-buffer "*Shell Command Output*")
    (shell-command
     (funcall *krb-lisp-unit-test-cmdline-func*))
    ;; TODO: how can I get that buffer to jump to the bottom so the unit
    ;; test output / summary is shown?  For some reason this executes
    ;; before the output is copied into that buffer so it has no effect...
    (set-buffer "*Shell Command Output*")
    (end-of-buffer)))


(defun krb-tmp ()
  (interactive)
  (message " => %s"
	   (krb-lisp-find-file-by-wildcard-traverse-up "*.asd")))


(defun krb-lisp-make-test-runner-cmdline ()
  "Generates a test runner command line for SBCL and lisp-unit.
This function currently assumes that you're creating and working
with an asdf defsystem'd library.  If you're using the functions
from this library to create and maintain your code, and you're
using sbcl (on your PATH), then you should be Ok."
  (interactive)
  (let* ((test-file-name (krb-lisp-get-test-file-name-for-current-buffer))
	 (base-file-name (krb-lisp-get-base-file-name-for-current-test-buffer test-file-name))
	 (test-package-name (krb-lisp-get-test-package-name-for-current-buffer))
	 (asdf-system-file (krb-lisp-find-file-by-wildcard-traverse-up "*.asd"))
	 (asdf-system-dir (file-name-directory asdf-system-file))
	 (asdf-system-name (concat ":" (file-name-sans-extension (file-name-nondirectory asdf-system-file)))))
    (concat 
     "sbcl --noinform"
     " --eval '(require :asdf)'"
     " --eval '(require :lisp-unit)'"
     (format  " --eval '(setf asdf:*central-registry* (cons \"%s\" asdf:*central-registry*))'"
	      asdf-system-dir)
     (format " --eval \"(asdf:oos 'asdf:load-op %s)\""
	     asdf-system-name)
     (format " --load %s " base-file-name)
     (format " --load %s " test-file-name)
     (format " --eval '(use-package :%s)'" test-package-name)
     (format " --eval '(progn (lisp-unit:use-debugger t) (lisp-unit:run-all-tests :%s))'"
	      test-package-name)
     " --eval '(sb-ext:quit)'")))

(defun krb-tmp ()
  (interactive)
  (message " => %s" (krb-lisp-make-test-runner-cmdline)))

(defun krb-tmp ()
  (interactive)
  (message " => %s" (funcall *krb-lisp-unit-test-cmdline-func*)))

;; TODO: what if it's not in the current buffer?  how to search?  Look
;; through all files in all the directories in the asdf central
;; registry?  What about using symbol-package?
;; TODO: can this fall back to M-. (the normal etags)?
(defun krb-lisp-jump-to-def-form-push (symbol-name)
  "Starting in the current buffer, searches for a
definition (defun, defmacro, defvar, etc) for the given symbol.
Jumps to the first instance found.  

If not found, it next searches all the lisp files in the same
directory as the file being edited."
  (interactive
   (list
    (read-string "Find Symbol: " (krb-lisp-escape-regex-characters (krb-lisp-find-likely-symbol)))))
  (let ((last-position (list (buffer-file-name)
			     (point)))
	(new-position nil))
    (condition-case
	()
	(setf new-position
	      (krb-lisp-perform-declaration-search (buffer-file-name) symbol-name))
      (error
       (loop for file in (file-expand-wildcards (concat (file-name-directory (buffer-file-name))
							"*.lisp"))
	     do
	     (message "trying: %s" file)
	     (setf new-position
		   (krb-lisp-perform-declaration-search file symbol-name)))))
    (if new-position
	(progn
	  (message "found, jumping to: %s" new-position)
	  (push last-position
		*krb-lisp-jump-stack*)
	  (find-file (first new-position))
	  (goto-char (second new-position))))))

(defun krb-lisp-pop-jump-stack ()
  "The author needs to document this function"
  (interactive)
  (let ((pair (pop *krb-lisp-jump-stack*)))
    (find-file (first pair))
    (goto-char (second pair))))

(defun krb-lisp-escape-regex-characters (str)
  "Attempts to make lisp symbols emacs regex safe (*)."
  (replace-regexp-in-string "\\([\\*]\\)" "\\\\\\1" str))

(defun krb-lisp-perform-declaration-search (file-name symbol-name)
  "Open the given file and look for the given declaration."
    (save-excursion
      (beginning-of-buffer)
      (search-forward-regexp (format "(def\\(class\\|constant\\|generic\\|in-condition\\|ine-symbol-macro\\|macro\\|method\\|package\\|parameter\\|setf\\|struct\\|type\\|un\\|var\\) %s " symbol-name))
      (list (buffer-file-name)
	    (point))))

;; TODO: from perl-utils, make this interactive so you can keep going
;; to the next instance, wrap around to the top, or go on to other
;; files...see `hms-perl-follow-stack-trace' as an example.
(defun krb-lisp-find-next-symbol-usage (symbol-name)
  "Looks for the next usage of the given symbol in the current buffer."
  (interactive
   (list
    (read-string "Symbol: " (krb-lisp-find-likely-symbol))))
  (search-forward symbol-name))


(defun krb-lisp-new-asdf-system (location package-name use-list)
  "Creates a new 'project', including a basic directory structure,"
  (interactive
   (list
    (read-string "Location: " *krb-lisp-lisp-project-directory*)
    (read-string "Package Name: " "")
    (read-string "Use Packages: " "")))
  (let ((pkg-location (format "%s/%s" location package-name)))
    (unless (file-directory-p pkg-location)
      (make-directory pkg-location t))
    (find-file (format "%s/%s.asd" pkg-location package-name))
    (krb-lisp-insert-defsystem package-name use-list)))


(defun krb-lisp-insert-defsystem (package-name use-list)
  "Generates the asdf defsystem declaration."
  (interactive "sPackage Name: \nsUse List: ")
  (message "creating : %s '%s'" package-name use-list)
  (krb-lisp-insert-indented-lines
   `("(asdf:defsystem #:" ,package-name "\n") 
   `(":depends-on ("
     ,(krb-lisp-join->string
       " "
       (mapcar #'(lambda (elt)
		   (format "\"%s\"" elt))
	       (split-string use-list))) ")\n")
   `(":components ((:file \"package\")\n")
   `("(:file \"" ,package-name "\"\n" ) 
   `(":depends-on (\"package\"))))\n")))

(defun krb-tmp ()
  (interactive)
  (message " => '%s'" (krb-lisp-join->string " " '(a b) "foo")))
